# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Data feeding module within the ketos library

    This module provides utilities to load data and feed it to models.

    Contents:
        BatchGenerator class
        TrainiDataProvider class
"""

import numpy as np
import pandas as pd
from sklearn.utils import shuffle
from ketos.data_handling.database_interface import parse_labels
from ketos.data_handling.data_handling import check_data_sanity, to1hot


class BatchGenerator():
    """ Creates batches to be fed to a model

        Instances of this class are python generators. They will load one batch at 
        a time from a HDF5 database, which is particularly useful when working with 
        larger than memory datasets. Yield (X,Y) or (ids,X,Y) if 'return_batch_ids' 
        is True. X is a batch of data as a np.array of shape (batch_size,mx,nx) where 
        mx,nx are the shape of on instance of X in the database. Similarly, Y is an 
        np.array of shape[0]=batch_size with the corresponding labels.

        It is also possible to load the entire data set into memory and provide it 
        to the BatchGenerator via the arguments x and y. This can be convenient when 
        working with smaller data sets.

        Args:
            batch_size: int
                The number of instances in each batch. The last batch of an epoch might 
                have fewer examples, depending on the number of instances in the hdf5_table.
            hdf5_table: pytables table (instance of table.Table()) 
                The HDF5 table containing the data
            x: numpy array
                Array containing the data images.
            y: numpy array
                Array containing the data labels.
            indices: list of ints
                Indices of those instances that will retrieved from the HDF5 table by the 
                BatchGenerator. By default all instances are retrieved.
            instance_function: function
                A function to be applied to the batch, transforming the instances. Must accept 
                'X' and 'Y' and, after processing, also return  'X' and 'Y' in a tuple.
            x_field: str
                The name of the column containing the X data in the hdf5_table
            y_field: str
                The name of the column containing the Y labels in the hdf5_table
            shuffle: bool
                If True, instances are selected randomly (without replacement). If False, 
                instances are selected in the order the appear in the database
            refresh_on_epoch: bool
                If True, and shuffle is also True, resampling is performed at the end of 
                each epoch resulting in different batches for every epoch. If False, the 
                same batches are used in all epochs.
                Has no effect if shuffle is False.
            return_batch_ids: bool
                If False, each batch will consist of X and Y. If True, the instance indices 
                (as they are in the hdf5_table) will be included ((ids, X, Y)).

        Attr:
            data: pytables table (instance of table.Table()) 
                The HDF5 table containing the data
            n_instances: int
                The number of intances (rows) in the hdf5_table
            n_batches: int
                The number of batches of size 'batch_size' for each epoch
            entry_indices:list of ints
                A list of all intance indices, in the order used to generate batches for this epoch
            batch_indices: list of tuples (int,int)
                A list of (start,end) indices for each batch. These indices refer to the 'entry_indices' attribute.
            batch_count: int
                The current batch within the epoch
            from_memory: bool
                True if the data are loaded from memory rather than an HDF5 table.

        Examples:
            >>> from tables import open_file
            >>> from ketos.data_handling.database_interface import open_table
            >>> h5 = open_file("ketos/tests/assets/15x_same_spec.h5", 'r') # create the database handle  
            >>> train_data = open_table(h5, "/train/species1")
            >>> train_generator = BatchGenerator(hdf5_table=train_data, batch_size=3, return_batch_ids=True) #create a batch generator 
            >>> #Run 2 epochs. 
            >>> n_epochs = 2    
            >>> for e in range(n_epochs):
            ...    for batch_num in range(train_generator.n_batches):
            ...        ids, batch_X, batch_Y = next(train_generator)   
            ...        print("epoch:{0}, batch {1} | instance ids:{2}, X batch shape: {3}, Y batch shape: {4}".format(e, batch_num, ids, batch_X.shape, batch_Y.shape))
            epoch:0, batch 0 | instance ids:[0, 1, 2], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:0, batch 1 | instance ids:[3, 4, 5], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:0, batch 2 | instance ids:[6, 7, 8], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:0, batch 3 | instance ids:[9, 10, 11], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:0, batch 4 | instance ids:[12, 13, 14], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:1, batch 0 | instance ids:[0, 1, 2], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:1, batch 1 | instance ids:[3, 4, 5], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:1, batch 2 | instance ids:[6, 7, 8], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:1, batch 3 | instance ids:[9, 10, 11], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            epoch:1, batch 4 | instance ids:[12, 13, 14], X batch shape: (3, 2413, 201), Y batch shape: (3,)
            >>> #Applying a custom function to the batch
            >>> #Takes the mean of each instance in X; leaves Y untouched
            >>> def apply_to_batch(X,Y):
            ...    X = np.mean(X, axis=(1,2)) #since X is a 3d array
            ...    return (X,Y)
            >>> train_generator = BatchGenerator(hdf5_table=train_data, batch_size=3, return_batch_ids=False, instance_function=apply_to_batch) 
            >>> X,Y = next(train_generator)                
            >>> #Now each X instance is one single number, instead of a (2413,201) matrix
            >>> #A batch of size 3 is an array of the 3 means
            >>> X.shape
            (3,)
            >>> #Here is how one X instance looks like
            >>> X[0]
            7694.1147
            >>> #Y is the same as before 
            >>> Y.shape
            (3,)
            >>> h5.close()
    """
    def __init__(self, batch_size, hdf5_table=None, x=None, y=None, indices=None, instance_function=None, x_field='data', y_field='boxes',\
                    shuffle=False, refresh_on_epoch_end=False, return_batch_ids=False):

        self.from_memory = x is not None and y is not None

        if self.from_memory:
            check_data_sanity(x, y)
            self.x = x
            self.y = y

            if indices is None:
                self.indices = np.arange(len(self.x), dtype=int) 
            else:
                self.indices = indices

            self.n_instances = len(self.indices)
        else:
            assert hdf5_table is not None, 'hdf5_table or x,y must be specified'
            self.data = hdf5_table
            self.x_field = x_field
            self.y_field = y_field
            if indices is None:
                self.n_instances = self.data.nrows
                self.indices = np.arange(self.n_instances)
            else:
                self.n_instances = len(indices)
                self.indices = indices

        self.batch_size = batch_size
        self.shuffle = shuffle
        self.instance_function = instance_function
        self.batch_count = 0
        self.refresh_on_epoch_end = refresh_on_epoch_end
        self.return_batch_ids = return_batch_ids

        self.n_batches = int(np.ceil(self.n_instances / self.batch_size))

        self.entry_indices = self.__update_indices__()

        self.batch_indices = self.__get_batch_indices__()

    
    def __update_indices__(self):
        """Updates the indices used to divide the instances into batches.

            A list of indices is kept in the self.entry_indices attribute.
            The order of the indices determines which instances will be placed in each batch.
            If the self.shuffle is True, the indices are randomly reorganized, resulting in batches with randomly selected instances.

            Returns
                indices: list of ints
                    The list of instance indices
        """
        indices = self.indices

        if self.shuffle:
            np.random.shuffle(indices)

        return indices

    def __get_batch_indices__(self):
        """Selects the indices for each batch

            Divides the instances into batchs of self.batch_size, based on the list generated by __update_indices__()

            Returns:
                list_of_indices: list of tuples
                    A list of tuple, each containing two integer values: the start and end of the batch. These positions refer to the list stored in self.entry_indices.                
        
        """
        ids = self.entry_indices
        n_complete_batches = int(self.n_instances // self.batch_size) # number of batches that can accomodate self.batch_size intances
        last_batch_size = self.n_instances % n_complete_batches
    
        list_of_indices = [list(ids[(i*self.batch_size):(i*self.batch_size)+self.batch_size]) for i in range(self.n_batches)]
        if last_batch_size > 0:
            last_batch_ids = list(ids[-last_batch_size:])
            list_of_indices.append(last_batch_ids)

        return list_of_indices

    def __iter__(self):
        return self

    def __next__(self):
        """         
            Return: tuple
            A batch of instances (X,Y) or, if 'returns_batch_ids" is True, a batch of instances accompanied by their indices (ids, X, Y) 
        """

        batch_ids = self.batch_indices[self.batch_count]

        if self.from_memory:
            X = np.take(self.x, batch_ids, axis=0)
            Y = np.take(self.y, batch_ids, axis=0)
        else:
            X = self.data[batch_ids][self.x_field]
            Y = self.data[batch_ids][self.y_field]

        self.batch_count += 1
        if self.batch_count > (self.n_batches - 1):
            self.batch_count = 0
            if self.refresh_on_epoch_end:
                self.entry_indices = self.__update_indices__()
                self.batch_indices = self.__get_batch_indices__()

        if self.instance_function is not None:
            X,Y = self.instance_function(X,Y)

        if self.return_batch_ids:
            return (batch_ids,X,Y)
        else:
            return (X, Y)

class ActiveLearningBatchGenerator():
    """ Creates batch generators to be used in active learning.

        The learning process is divided into sessions. During each 
        session the neural network considers only a part of the 
        entire training data set. Typically, the network is exposed 
        to a mixture of new examples and examples from the previous 
        session, which it failed to learn (or learned, but with low 
        confidence).

        Args:
            session_size: int
                The number of instances in each learning session.
            batch_size: int
                The number of instances in each batch. The last batch of 
                an epoch might have fewer examples, depending on the number 
                of instances in the hdf5_table.
            num_labels: int
                Number of labels
            hdf5_table: pytables table (instance of table.Table()) 
                The HDF5 table containing the data
            x: numpy array
                Array containing the data images.
            y: numpy array
                Array containing the data labels.
            instance_function: function
                A function to be applied to the batch, transforming the instances. 
                Must accept 'X' and 'Y' and, after processing, also return  'X' and 
                'Y' in a tuple.
            x_field: str
                The name of the column containing the X data in the hdf5_table
            y_field: str
                The name of the column containing the Y labels in the hdf5_table
            shuffle: bool
                If True, instances are selected randomly (without replacement). If 
                False, instances are selected in the order the appear in the database
            refresh: bool
                If True, and shuffle is also True, resampling is performed at the end of 
                each epoch resulting in different batches for every epoch. If False, the 
                same batches are used in all epochs. Has no effect if shuffle is False.
            return_indices: bool
                If False, each batch will consist of X and Y. If True, the instance indices 
                (as they are in the hdf5_table) will be included ((ids, X, Y)).
            max_keep: float
                Maximum number of examples that are kept from the previous session, expressed 
                as a fraction of the session size
            conf_cut: float
                Correct predictions with confidence below conf_cut will be kept for next session 
                (all wrong predictions are also kept)
            seed: int
                Seed for random number generator
            parse_labels: bool
                Parse labels field. Only applicable if y_field is 'labels' and instance_function is None.
            convert_to_one_hot: bool
                Convert labels to 1-hot representation. Only applicable if instance_function is None.
            val_frac: float between 0 and 1
                Fraction of data that will be used for validation
            val_size: int
                Number of samples that will be used for validation. 
                None by default. Overwrites val_frac.
            suppress_warnings: bool
                Do not show warnings

        Attr:
            data: pytables table (instance of table.Table()) 
                The HDF5 table containing the data
            data_size: int
                The total number of instances in the data set
            indices: numpy array
                Indices of all instances in the data set
            session_indices: numpy array  
                Indices of all instances in the current session              
            poor_indices: numpy array
                Indices of those instances in the previous session that were wronly 
                predicted, or correctly predicted but with low confidence
            from_memory: bool
                True if the data are loaded from memory rather than an HDF5 table.
            suppress_warnings: bool
                Do not show warnings
            val_frac: float between 0 and 1
                Fraction of data that will be used for validation

        Example:
            >>> from tables import open_file
            >>> from ketos.data_handling.database_interface import open_table
            >>> h5 = open_file("ketos/tests/assets/15x_same_spec.h5", 'r') # open the database file 
            >>> table = open_table(h5, "/train/species1") # access the table of interest
            >>> # create an active learning module with session_size of 6, batch_size of 3 and max_keep of 0.4
            >>> active_learning = ActiveLearningBatchGenerator(session_size=6, batch_size=3, table=table, max_keep=0.4, seed=1, return_indices=True) 
            >>> # run 2 sessions, and 3 epochs for each session
            >>> num_sessions = 2
            >>> num_epochs = 3  
            >>> for s in range(num_sessions): # loop over sessions
            ...    generator = next(active_learning)
            ...    for e in range(num_epochs): # loop over epochs
            ...       for b in range(generator.n_batches): # loop over batches
            ...          idx, X, Y = next(generator)
            ...          # pretend that neural network makes predictions [1 1 0]
            ...          active_learning.update_performance(idx, predictions=[1, 1, 0])  
            ...          print("session:{0}, epoch:{1}, batch:{2} | instance: {3}, X shape: {4}, Y length: {5}".format(s, e, b, idx, X.shape, len(Y)))
            session:0, epoch:0, batch:0 | instance: [0, 1, 2], X shape: (3, 2413, 201), Y length: 3
            session:0, epoch:0, batch:1 | instance: [3, 4, 5], X shape: (3, 2413, 201), Y length: 3
            session:0, epoch:1, batch:0 | instance: [0, 1, 2], X shape: (3, 2413, 201), Y length: 3
            session:0, epoch:1, batch:1 | instance: [3, 4, 5], X shape: (3, 2413, 201), Y length: 3
            session:0, epoch:2, batch:0 | instance: [0, 1, 2], X shape: (3, 2413, 201), Y length: 3
            session:0, epoch:2, batch:1 | instance: [3, 4, 5], X shape: (3, 2413, 201), Y length: 3
            session:1, epoch:0, batch:0 | instance: [5, 8, 7], X shape: (3, 2413, 201), Y length: 3
            session:1, epoch:0, batch:1 | instance: [6, 2, 9], X shape: (3, 2413, 201), Y length: 3
            session:1, epoch:1, batch:0 | instance: [5, 8, 7], X shape: (3, 2413, 201), Y length: 3
            session:1, epoch:1, batch:1 | instance: [6, 2, 9], X shape: (3, 2413, 201), Y length: 3
            session:1, epoch:2, batch:0 | instance: [5, 8, 7], X shape: (3, 2413, 201), Y length: 3
            session:1, epoch:2, batch:1 | instance: [6, 2, 9], X shape: (3, 2413, 201), Y length: 3
            >>> h5.close()
    """
    def __init__(self, session_size, batch_size, num_labels=2, table=None, x=None, y=None, shuffle=False, refresh=False, return_indices=False,\
                    max_keep=0, conf_cut=0, seed=None, instance_function=None, x_field='data', y_field='labels', parse_labels=True,\
                    convert_to_one_hot=False, val_frac=0.0, val_size=None, suppress_warnings=False):

        self.from_memory = x is not None and y is not None

        if self.from_memory:
            check_data_sanity(x, y)
            self.x = x
            self.y = y
            self.data_size = len(x)
            self.data = None
            self.x_field = None
            self.y_field = None
        else:
            assert table is not None, 'table or x,y must be specified'
            self.data = table
            self.data_size = self.data.nrows
            self.x_field = x_field
            self.y_field = y_field
            self.x = None
            self.y = None

        self.session_size = session_size
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.refresh = refresh
        self.return_indices = return_indices
        self.max_keep = max_keep
        self.conf_cut = conf_cut
        self.seed = seed
        self.instance_function = instance_function
        self.num_labels = num_labels
        self.convert_to_one_hot = convert_to_one_hot
        self.parse_labels = parse_labels
        self.instance_function = instance_function
        self.suppress_warnings = suppress_warnings

        if seed is not None:
            np.random.seed(seed) 

        self.poor_indices = np.array([], dtype=int)
        self.indices = np.arange(self.data_size, dtype=int) 

        # reserve for validation
        self.val_indices = []
        if val_size is not None:
            val_frac = min(1.0, val_size / self.data_size)

        if val_frac > 0:
            num_val = max(1, int(val_frac * self.data_size))
            self.val_indices = np.random.choice(self.indices, num_val, replace=False)
            self.indices = np.delete(self.indices, self.val_indices)

        self.indices = self.__refresh_indices__()

        self.counter = 0


    def __instance_function__(self):
        """Creates the instance function for the batch generator

            Returns:
                func: function
                    The instance function               
        
        """
        if self.instance_function:
            self.__f1__ = self.instance_function
            self.__f2__ = self.__identity__
            self.__f3__ = self.__identity__
        else:
            self.__f1__ = self.__identity__
            if self.parse_labels and self.y_field == 'labels':
                self.__f2__ = self.__parse_labels__
            else:
                self.__f2__ = self.__identity__
            if self.convert_to_one_hot:
                self.__f3__ = self.__convert_to_one_hot__
            else:
                self.__f3__ = self.__identity__

        def func(X,Y):
            X,Y = self.__f1__(X,Y)
            X,Y = self.__f2__(X,Y)
            X,Y = self.__f3__(X,Y)
            return X,Y

        return func

    def __refresh_indices__(self):
        """Updates the indices used to divide the data into sessions

            The order of the indices determines which instances will be placed in each batch.
            If the self.shuffle is True, the indices are randomly reorganized, resulting in batches with randomly selected instances.

            Returns
                indices: list of ints
                    The list of instance indices
        """
        indices = self.indices

        if self.shuffle:
            np.random.shuffle(indices)

        return indices

    def __get_session_indices__(self):
        """Selects the indices for the next session

            Returns:
                list_of_indices: list of ints
                    Indices for the next session                
        
        """
        # number of instances from previous session with poor performace 
        num_poor = len(np.unique(self.poor_indices))

        # number of examples kept from previous session
        num_keep = int(min(num_poor, self.max_keep * self.session_size))

        # number of new examples
        num_new = self.session_size - num_keep

        # select new examples
        i1 = self.counter
        i2 = min(i1 + num_new, self.data_size)
        new_session = self.indices[i1:i2]
        self.counter = i2

        # if necessary, go back to beginning to complete batch
        dn = num_new - new_session.shape[0]
        if dn > 0:
            new_session = np.concatenate((new_session, self.indices[:dn]))
            self.counter = dn

        # select randomly from poorly predicted examples in previous session
        if num_keep > 0:
            new_session = np.concatenate((new_session, np.random.choice(np.unique(self.poor_indices), num_keep, replace=False)))

        # refresh at end of data set
        epoch_end = (i2 == self.data_size or dn > 0)
        if self.refresh and epoch_end:
            self.indices = self.__refresh_indices__()

        # shuffle new session, if it contains examples from previous session
        if num_keep > 0:
            np.random.shuffle(new_session)

        return new_session

    def __iter__(self):
        return self

    def __next__(self):
        """         
            Return: BatchGenerator
                Batch generator for the next session.
        """
        # get indices for new session
        session_indices = self.__get_session_indices__()

        # refresh poor_indices array
        self.poor_indices = np.array([], dtype=int)

        # create batch generator
        generator = BatchGenerator(hdf5_table=self.data, x=self.x, y=self.y, batch_size=self.batch_size, indices=session_indices,\
                    instance_function=self.__instance_function__(), x_field=self.x_field, y_field=self.y_field,\
                    shuffle=self.shuffle, refresh_on_epoch_end=self.refresh, return_batch_ids=self.return_indices)

        return generator

    def get_validation_data(self):
        """ Get validation data.

                Returns: tuple
                    Instances (X,Y) or, if 'return_indices" is True, instances accompanied by their indices (ids, X, Y) 
        """
        if len(self.val_indices) == 0:
            X, Y = [], []
    
        else:
            if self.from_memory:
                X = np.take(self.x, self.val_indices, axis=0)
                Y = np.take(self.y, self.val_indices, axis=0)
            else:
                X = self.data[self.val_indices][self.x_field]
                Y = self.data[self.val_indices][self.y_field]

            X, Y = self.__instance_function__()(X, Y)

        if self.return_indices:
            return (self.val_indices, X, Y)
        else:
            return (X, Y)        

    def update_performance(self, indices, predictions, confidences=None):
        """Inform the generator about how well the neural network performed on a set of examples.

            Args:
                indices: numpy.array
                    Indices of the examples that the predictions and confidences refer to
                pred: numpy.array
                    Array containing the predictions on the last batch.
                    Created with ketos.neural_networks.neural_networks.predictions.
                conf: numpy.array
                    Array containing the confidences for the predictions on the last batch.
                    Created with ketos.neural_networks.neural_networks.class_confidences.    
        """
        if confidences is None:
            confidences = np.ones(len(predictions))
            
        assert len(indices) == len(predictions), 'length of indices and predictions arrays do not match'
        assert len(predictions) == len(confidences), 'length of prediction and confidence arrays do not match'

        if type(indices) != np.ndarray:
            indices = np.array(indices, dtype=int)

        if type(predictions) != np.ndarray:
            predictions = np.array(predictions, dtype=int)

        if type(confidences) != np.ndarray:
            confidences = np.array(confidences, dtype=float)

        if self.from_memory:
            Y = np.take(self.y, indices)            
        else:
            Y = self.data[indices][self.y_field]

        if self.parse_labels and self.y_field == 'labels':
            _, Y = self.__parse_labels__(None, Y)

        poor = np.logical_or(predictions != Y, confidences < self.conf_cut)
        poor = np.argwhere(poor == True)
        poor = np.squeeze(poor)
        if np.ndim(poor) == 0:
            poor = np.array([poor], dtype=int)

        self.poor_indices = np.concatenate((self.poor_indices, indices[poor]))

    def __identity__(self,X,Y):
        """Identity function.
        """
        return X,Y 

    def __parse_labels__(self,X,Y):
        """Parse labels extracted from hdf5 database.
        """
        YY = list()
        for _y in Y:
            _y = parse_labels(_y)
            if len(_y) >= 1:# if there are one or more labels, we use the first label
                if len(_y) > 1:
                    if not self.suppress_warnings:
                        print('Warning: ActiveLearningBatchGenerator encountered instance with more than one label. Using the first label only.')

                _y = _y[0]
            elif len(_y) == 0:# if there are no labels, we use the label 0
                _y = 0

            YY.append(_y)

        if len(YY) == 1:
            YY = YY[0]        

        return X,YY

    def __convert_to_one_hot__(self,X,Y):
        """Convert labels to 1-hot encoding
        """
        Y = to1hot(Y, self.num_labels)
        return X,Y




# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'neural_networks' module within the ketos library
"""

import pytest
import numpy as np
import pandas as pd
import ketos.data_handling.data_handling as dh
from ketos.neural_networks.neural_networks import DataHandler, DataUse, class_confidences
from tensorflow import reset_default_graph


@pytest.mark.test_DataHandler
def test_initialize_DataHandler(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    x = d["train_x"]
    y = d["train_y"]
    print(x.shape)
    print(y.shape)
    _ = DataHandler(train_x=x, train_y=y)

@pytest.mark.test_DataHandler
def test_set_data_to_DataHandler(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    x = d["train_x"]
    y = d["train_y"]
    network = DataHandler(train_x=x, train_y=y)
    # set training data
    network.set_training_data(x=x, y=y)
    assert np.all(x == network.images[DataUse.TRAINING])
    assert np.all(y == network.labels[DataUse.TRAINING])
    # set validation data
    network.set_validation_data(x=x, y=y)
    assert np.all(x == network.images[DataUse.VALIDATION])
    assert np.all(y == network.labels[DataUse.VALIDATION])
    # set test data
    network.set_test_data(x=x, y=y)
    assert np.all(x == network.images[DataUse.TEST])
    assert np.all(y == network.labels[DataUse.TEST])    

@pytest.mark.test_DataHandler
def test_add_data_to_DataHandler(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    x = d["train_x"]
    y = d["train_y"]
    network = DataHandler(train_x=x, train_y=y)
    # add training data
    network.add_training_data(x=x, y=y)
    assert 2 * x.shape[0] == network.images[DataUse.TRAINING].shape[0]
    assert x.shape[1:] == network.images[DataUse.TRAINING].shape[1:]
    assert 2 * y.shape[0] == network.labels[DataUse.TRAINING].shape[0]
    assert y.shape[1:] == network.labels[DataUse.TRAINING].shape[1:]
    
@pytest.mark.test_DataHandler
def test_get_data_from_DataHandler(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    x = d["train_x"]
    y = d["train_y"]
    network = DataHandler(train_x=x, train_y=y)
    # get training data
    xx, yy = network.get_training_data()
    assert x.shape == xx.shape
    assert y.shape == yy.shape

@pytest.mark.test_class_confidences
def test_class_confidences(data_classified_by_nn):
    _,_,w = data_classified_by_nn
    conf = class_confidences(class_weights=w)
    assert len(conf) == 6
    assert conf[0] == pytest.approx(0.6, abs=0.001)
    assert conf[1] == pytest.approx(0.8, abs=0.001)



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" spectrogram_filters module within the ketos library

    This module contains various filters that can be applied to spectrograms.

    Filters are implemented as classes, which are required to have the 
    attribute name and the method apply. 

    Contents:
        FAVThresholdFilter class
        FAVFilter class
        HarmonicFilter class
        FrequencyFilter class
        WindowFilter class
        WindowSubtractionFilter class
        AverageFilter class
"""

import numpy as np
import pandas as pd
from ketos.audio_processing.spectrogram import Spectrogram
from ketos.utils import nearest_values



class FAVThresholdFilter():
    """ Modified Frequency Amplitude Variation (FAV) filter 
        combined with thresholding operation.

        Args:
            threshold: float
                Peak detection threshold 
            winlen: int
                Length of smoothing window
        
        Attributes:
            threshold: float
                Peak detection threshold, in units of std. dev.
            winlen: int
                Length of smoothing window
            name: str
                Filter name
            std: numpy array
                Standard deviation of the smoothened spectrum 
    """

    def __init__(self, threshold=3.0, winlen=7):

        self.threshold = threshold
        self.winlen = winlen
        self.name = "FAV"
        self.std = None
        self.FAV = FAVFilter(winlen)

    def apply(self, spec):
        """ Apply FAV thresholding filter to spectrogram

            Args:
                spec: Spectrogram
                    Spectrogram to which filter will be applied 
        """
        f = self.FAV
        N = self.winlen
        orig = spec.copy()

        # apply FAV filter
        f.apply(spec)

        z = np.empty((spec.image.shape[0],2))

        # loop over 1st dimension
        m = spec.image.shape[0]
        for i in range(m):
            x = spec.image[i]

            # thresholding operation
            y = x * (x > self.threshold * f.std[i])

            # find peaks
            y1 = np.r_[y[1:], 0]
            y2 = np.r_[0, y[:-1]]
            peaks = np.argwhere(np.logical_and(y > y1, y >= y2))
            peaks = np.squeeze(peaks)
            if np.ndim(peaks) == 0:
                peaks = np.array([peaks])

            # number of peaks and max strength
            num_peaks = len(peaks)
            max_strength = 0
            for p1 in peaks:
                p2 = p1 + max(2, int(N/2))
                max_strength = max(max_strength, np.max(orig.image[i,p1:p2]))

            z[i,0] = num_peaks
            z[i,1] = max_strength / 100.                                   

        spec.image = z
        spec.flabels = ['num_peaks', 'max_peak_strength']


class FAVFilter():
    """ Modified Frequency Amplitude Variation (FAV) filter.

        Algorithm based on Reis et al. "Automatic Detection of Vessel 
        Signatures in Audio Recordings with Spectral Amplitude Variation 
        Signature" (2019).

        First, the spectrogram is smoothened along the frequency axis 
        using a Blackman filter.
    
        Second, the difference between neighboring bins is computed, 
        also along the frequency axis, and raised to the 3rd power.        

        Third, the resulting matrix is shifted by 3 bins, along the 
        frequency axis, inverted, and multiplied with the unshifted matrix. 

        Finally, only positive values are retained.

        The result of this set of operations is to emphasize narrow 
        peaks in the frequency spectrum. (Here narrow implies narrower than 
        roughly half of the Blackman filter.)

        Args:
            winlen: int
                Length of smoothing window
        
        Attributes:
            winlen: int
                Length of smoothing window
            name: str
                Filter name
            std: numpy array
                Standard deviation of the smoothened spectrum 
    """

    def __init__(self, winlen=7):

        self.winlen = winlen
        self.name = "FAV"
        self.std = None

    def apply(self, spec):
        """ Apply FAV filter to spectrogram

            Args:
                spec: Spectrogram
                    Spectrogram to which filter will be applied 
        """
        x = spec.image
        N = self.winlen
        D = x.ndim

        assert D == 2, "FAVFilter.apply only accepts 2 dimensional arrays."
        assert x.shape[1] > N, "Frequency axis must be longer than window size."

        if N > 1:        
            # pad with reflected copies
            x = np.swapaxes(x,0,1)
            s = np.r_[x[N-1:0:-1], x, x[-2:-N-1:-1]]  
            x = np.swapaxes(x,0,1)        
            s = np.swapaxes(s,0,1)

            # blackman averaging filter
            w = np.blackman(N)

            # convolution
            K = x.shape[0]
            M = x.shape[1]
            y = np.empty((K,M+N-1))
            for i in range(s.shape[0]):
                y[i] = np.convolve(w/w.sum(), s[i], mode='valid')

            # remove ends
            k1 = int(N/2)
            k2 = -k1 + 2 - N%2
            x = y[:,k1:k2]

        # compute standard deviation
        self.std = np.std(x, axis=1)

        # difference to power of 3
        x = np.power(np.diff(x, axis=1), 3)

        # shift, invert and multiply by itself
        if N > 1:
            L = int(N/2)
        else:
            L = 1
        y = -x[:, L:]
        y = np.concatenate((y,np.zeros((y.shape[0],L))), axis=1)
        x = x * y
        x = x * (x > 0)

        spec.image = x


class HarmonicFilter():
    """ Performs Fast Fourier Transform (FFT) of the frequency axis.

        Attributes:
            name: str
                Filter name
    """

    def __init__(self):

        self.name = "Harmonic"

    def apply(self, spec):
        """ Apply harmonic filter to spectrogram

            Args:
                spec: Spectrogram
                    Spectrogram to which filter will be applied 
        """
        h = np.abs(np.fft.rfft(spec.image, axis=1))
        new_img = h[:,1:]

        flabels = list()
        for i in range(new_img.shape[1]):
            f = 2 * float(i) / float(new_img.shape[1]) * (spec.fmax() - spec.fmin)
            flabels.append("{0:.1f}Hz".format(f))

        spec.image = new_img
        spec.flabels = flabels


class CroppingFilter():
    """ Crops along the frequency dimension.

        Args:
            flow: float
                Lower bound on frequency in Hz
            fhigh: float
                Upper bound on frequency in Hz

        Attributes:
            name: str
                Filter name
    """

    def __init__(self, flow=None, fhigh=None):

        self.flow = flow
        self.fhigh = fhigh
        self.name = "Cropping"

    def apply(self, spec):
        """ Apply frequency cropping filter to spectrogram

            Args:
                spec: Spectrogram
                    Spectrogram to which filter will be applied 
        """
        spec.crop(flow=self.flow, fhigh=self.fhigh)


class FrequencyFilter():
    """ Splits the frequency axis into bands and computes 
        the average sound magnitude within each band.

        Args:
            bands: list(Interval)
                Frequency bands in Hz
            names: list(str)
                Names of the frequency bands

        Attributes:
            name: str
                Filter name
    """

    def __init__(self, bands, names=None):

        self.bands = bands
        self.names = names
        self.name = "Frequency"

    def apply(self, spec):
        """ Apply frequency filter to spectrogram

            Args:
                spec: Spectrogram
                    Spectrogram to which filter will be applied 
        """
        nt = spec.tbins()
        nf = len(self.bands)

        new_img = np.ma.zeros(shape=(nt,nf))

        for i in range(nf):
            b = self.bands[i]
            new_img[:,i] = spec.average(axis=1, flow=b.low, fhigh=b.high)

        # discard frequency bands with NaN's
        not_nan = ~np.any(np.isnan(new_img), axis=0)
        new_img = new_img[:, not_nan]
        names = [i for (i, v) in zip(self.names, not_nan) if v]

        # mask zeros
        new_img = np.ma.masked_values(new_img, 0)

        spec.image = new_img
        spec.flabels = names


class WindowFilter():
    """ Applies a windowed median/average filter to the spectrogram.

        Args:
            window_size: float
                Window size in seconds
            step_size: float
                Step size in seconds
            filter_func: 
                Filtering function. Default is np.ma.median

        Attributes:
            name: str
                Filter name
    """

    def __init__(self, window_size, step_size, filter_func=np.ma.median):

        self.window_size = window_size
        self.step_size = step_size
        self.filter_func = filter_func

        if filter_func is np.ma.median:
            self.name = "Median"
        elif filter_func is np.ma.average:
            self.name = "Average"
        else:
            self.name = "Window"
            

    def apply(self, spec):
        """ Apply filter to spectrogram

            Args:
                spec: Spectrogram
                    Spectrogram to which filter will be applied 
        """
        step = int(np.ceil(self.step_size / spec.tres))
        window = int(np.ceil(self.window_size / spec.tres))
        
        nt = spec.tbins()
        nf = spec.fbins()
        
        n = int(np.ceil(nt/step))

        img = spec.image
        new_img = np.zeros(shape=(n,nf))

        for i in range(n):
            i1 = i * step
            i2 = min(nt-1, i1 + window)
            new_img[i,:] = self.filter_func(img[i1:i2+1,:], axis=0)  # ignore entries with value=0

        # mask zeros
        new_img = np.ma.masked_values(new_img, 0)    

        spec.image = new_img
        spec.tres = step * spec.tres


class WindowSubtractionFilter():
    """ Applied a windowed median/average subtraction filter to the spectrogram.

        Args:
            window_size: float
                Window size in seconds
            filter_func: 
                Filtering function. Default is np.ma.median

        Attributes:
            name: str
                Filter name
    """

    def __init__(self, window_size, filter_func=np.ma.median):

        self.window_size = window_size
        self.filter_func = filter_func

        if filter_func is np.ma.median:
            self.name = "Median Subtraction"
        elif filter_func is np.ma.average:
            self.name = "Average Subtraction"
        else:
            self.name = "Window Subtraction"

    def apply(self, spec):
        """ Apply filter to spectrogram

            Args:
                spec: Spectrogram
                    Spectrogram to which filter will be applied 
        """
        window = 1 + 2 * int(0.5 * self.window_size / spec.tres)
        
        nt = spec.tbins()
        nf = spec.fbins()

        img = spec.image
        new_img = np.zeros(shape=(nt,nf))

        # loop over time bins
        for i in range(nt):

            # local median
            v = nearest_values(x=img, i=i, n=window)
            med = self.filter_func(v, axis=0)
            
            new_img[i,:] = img[i,:] - med
        
        spec.image = new_img


class AverageFilter(WindowFilter):
    """ Applies a windowed average filter to the spectrogram.

        Args:
            window_size: float
                Window size in seconds
            step_size: float
                Step size in seconds

        Attributes:
            name: str
                Filter name
    """
    def __init__(self, window_size, step_size):
        super().__init__(window_size, step_size, np.ma.average)


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'parsing' module within the ketos library
"""

import pytest
import json
import ketos.data_handling.parsing as jp

@pytest.fixture
def spectr_config_json_complete():
    j = '{"spectrogram": {"rate": "20 kHz", "window_size": "0.1 s", "step_size": "0.025 s", "bins_per_octave": "32",\
        "window_function": "HAMMING", "low_frequency_cut": "30Hz", "high_frequency_cut": "3000Hz",\
        "length": "1.0s", "overlap": "0.2s", "type": "Mag"}}'
    return j

@pytest.fixture
def spectr_config_json_partial():
    j = '{"spectrogram": {"rate": "20 kHz", "window_size": "0.1 s"}}'
    return j

@pytest.fixture
def one_frequency_band_json():
    j = '{"frequency_bands": [{"name": "15.6Hz", "range": ["11.0Hz", "22.1Hz"]}]}'
    return j

@pytest.fixture
def two_frequency_bands_json():
    j = '{"frequency_bands": [{"name": "15.6Hz", "range": ["11.0Hz", "22.1Hz"]},{"name": "test", "range": ["9kHz", "10kHz"]}]}'
    return j

@pytest.fixture
def frequency_band_without_name_json():
    j = '{"frequency_bands": [{"range": ["11.0Hz", "22.1Hz"]}]}'
    return j

@pytest.mark.test_parse_spectrogram_configuration
def test_parse_complete_spectrogram_config(spectr_config_json_complete):
    data = json.loads(spectr_config_json_complete)
    cfg = jp.parse_spectrogram_configuration(data['spectrogram'])
    assert cfg.rate == 20000
    assert cfg.window_size == 0.1
    assert cfg.step_size == 0.025
    assert cfg.bins_per_octave == 32
    assert cfg.window_function == jp.WinFun.HAMMING
    assert cfg.low_frequency_cut == 30
    assert cfg.high_frequency_cut == 3000
    assert cfg.length == 1.0
    assert cfg.overlap == 0.2
    assert cfg.type == 'Mag'

@pytest.mark.test_parse_spectrogram_configuration
def test_parse_partial_spectrogram_config(spectr_config_json_partial):
    data = json.loads(spectr_config_json_partial)
    cfg = jp.parse_spectrogram_configuration(data['spectrogram'])
    assert cfg.rate == 20000
    assert cfg.window_size == 0.1
    assert cfg.bins_per_octave == None
    assert cfg.step_size == None
    assert cfg.window_function == None    
    assert cfg.length == None
    assert cfg.overlap == None
    assert cfg.type == None

@pytest.mark.test_parse_frequency_bands
def test_parse_one_frequency_band(one_frequency_band_json):
    data = json.loads(one_frequency_band_json)
    band_name, freq_interval = jp.parse_frequency_bands(data['frequency_bands'])
    assert len(band_name) == 1
    assert band_name[0] == '15.6Hz'
    assert freq_interval[0].low == 11.0
    assert freq_interval[0].high == 22.1

@pytest.mark.test_parse_frequency_bands
def test_parse_two_frequency_bands(two_frequency_bands_json):
    data = json.loads(two_frequency_bands_json)
    band_name, freq_interval = jp.parse_frequency_bands(data['frequency_bands'])
    assert len(band_name) == 2
    assert band_name[1] == 'test'
    assert freq_interval[1].low == 9000.0
    assert freq_interval[1].high == 10000.0

@pytest.mark.test_parse_frequency_bands
def test_parse_frequency_band_without_name(frequency_band_without_name_json):
    data = json.loads(frequency_band_without_name_json)
    with pytest.raises(AssertionError):
        _, _ = jp.parse_frequency_bands(data['frequency_bands'])


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Annotation module within the ketos library

    This module provides utilities to handle annotations associated 
    with audio files and spectrograms.

    Contents:
        AnnotationHandler class: 
"""

from ketos.utils import ndim
import numpy as np
import math


class AnnotationHandler():
    """ Parent class for the AudioSignal and Spectrogram classes.

        An annotation consists of an integer label (0,1,2,...) and 
        a bounding box that delimits the audio segment of interest 
        in time and optionally frequency. 
        
        The bounding box is a list of four floats: start time, end time,
        minimum frequency, and maximum frequency (with frequency expressed in Hz)
        A list of two floats ([start time, end time]) can also be used in the initialization.
        In this case, the minimum and maximum frequencies will be respectivelly filled as 0 and Inf.
        is interpreted as the start and end time of the 
        segment (in seconds). list of four floats is interpreted as 
        .

        Args:
            labels: list(int)
                List of annotation labels
            boxes: list(tuple)
                List of bounding boxes, each specifying a time interval (in seconds)
                and optionally also a frequency interval (in Hz).
                The format is [t_min, t_max, f_min, f_max]

        Attributes:
            labels: list(int)
                List of annotation labels
            boxes: list(tuple)
                List of bounding boxes, each specifying a time interval (in seconds)
                and optionally also a frequency interval (in Hz).
                The format is [t_min, t_max, f_min, f_max]
            precision: int
                Number of decimals used for bounding box values
    """
    def __init__(self, labels=None, boxes=None):

        self.labels = []
        self.boxes = []
        self.precision = 3

        if labels is not None:
            self.annotate(labels, boxes)

    def annotate(self, labels, boxes):
        """ Add a set of annotations.
            
            Args:
                labels: list(int)
                    Annotation labels.
                boxes: list(tuple)
                    Annotation boxes, specifying the start and stop time of the annotation 
                    and, optionally, the minimum and maximum frequency.

            Example:
                >>> from ketos.audio_processing.annotation import AnnotationHandler
                >>> 
                >>> handler = AnnotationHandler()
                >>> labels = [0, 1]
                >>> boxes = [[10.0, 12.2, 110., 700.],[30., 34.]]
                >>> handler.annotate(labels, boxes)
                >>> print(handler.labels)
                [0, 1]
                >>> print(handler.boxes)
                [[10.0, 12.2, 110.0, 700.0], [30.0, 34.0, 0, inf]]
        """
        if ndim(labels) == 0:
            labels = [labels]

        if ndim(boxes) == 1 and len(boxes) > 0:
            boxes = [boxes]

        assert len(labels) == len(boxes), 'number of labels must equal number of boxes'
        assert ndim(labels) == 1, 'labels list has invalid dimension'
        assert ndim(boxes) == 2 or (ndim(boxes) == 1 and len(boxes) == 0), 'boxes list has invalid dimension'

        for l,b in zip(labels, boxes):
            b = self._ensure4D(b)
            self.labels.append(l)
            self.boxes.append(b)

        for b in self.boxes:
            b = np.array(b).tolist()

    def _ensure4D(self, b):
        """ Ensure that the annotation box has four entries.

            Set the minimum frequency to 0 and the maximum 
            frequency to infinity, if these are missing.
            
            Args:
                b: list or tuple
                   Bounding box with two or four entries 
        """
        for v in b:
            v = round(v, self.precision)

        if len(b) == 2:
            b = [b[0], b[1], 0, math.inf]
        
        assert len(b) == 4, 'Found box with {0} entries; all boxes must have either 2 or 4 entries'.format(len(b))
        
        return b

    def delete_annotations(self, id=None):
        """ Delete the annotation with the specified ID(s).

            If no ID is spefied, all annotations are deleted.
            
            Args:
                id: int or list(int)
                    ID of the annotation to be deleted.

            Example:
                >>> from ketos.audio_processing.annotation import AnnotationHandler
                >>> 
                >>> labels = [0, 1]
                >>> boxes = [[10.0, 12.2, 110., 700.],[30., 34.]]
                >>> handler = AnnotationHandler(labels, boxes)
                >>> handler.delete_annotations(0)
                >>> print(handler.labels)
                [1]
                >>> print(handler.boxes)
                [[30.0, 34.0, 0, inf]]
        """
        if id is None:
            self.labels = []
            self.boxes = []
        else:
            if ndim(id) == 0:
                id = [id]
            # sort id's in ascending order 
            id = sorted(id, reverse=True)
            # loop over id's
            for i in id:
                if i < len(self.labels):
                    # delete specified annotations
                    del self.labels[i]
                    del self.boxes[i]

    def get_cropped_annotations(self, t1=0, t2=math.inf, f1=0, f2=math.inf):
        """ Update boundary boxes in response to cropping operation in time and/or frequency.
            
            Args:
                t1: float
                    Lower time cut in seconds
                t2: float
                    Upper time cut in seconds
                f1: float
                    Lower frequency cut in Hz
                f2: float
                    Upper frequency in in Hz

            Example:
                >>> from ketos.audio_processing.annotation import AnnotationHandler
                >>> 
                >>> labels = [0, 1]
                >>> boxes = [[10.0, 12.2, 110., 700.],[30., 34.]]
                >>> handler = AnnotationHandler(labels, boxes)
                >>> cropped_labels, cropped_boxes = handler.get_cropped_annotations(t1=10.3, f2=555.)
                >>> print(cropped_labels)
                [0, 1]
                >>> print(cropped_boxes)
                [[10.3, 12.2, 110.0, 555.0], [30.0, 34.0, 0, 555.0]]
        """
        if t1 is None: t1 = 0
        if t2 is None: t2 = math.inf
        if f1 is None: f1 = 0
        if f2 is None: f2 = math.inf

        labels, boxes = [], []

        # loop over annotations
        for l, b in zip(self.labels, self.boxes):

            # check if box overlaps with cut
            if b[0] < t2 and b[1] > t1 and b[2] < f2 and b[3] > f1:

                # update box boundaries
                b0 = max(b[0], t1)
                b1 = min(b[1], t2)
                b2 = max(b[2], f1)
                b3 = min(b[3], f2)

                # truncate to desired precision
                b0 = round(b0, self.precision)
                b1 = round(b1, self.precision)
                b2 = round(b2, self.precision)
                b3 = round(b3, self.precision)

                # new bounding box
                box = [b0, b1, b2, b3]

                labels.append(l)
                boxes.append(box)

        return labels, boxes

    def _shift_annotations(self, delay=0):
        """ Shift all annotations by a fixed time interval.
            
            Args:
                delay: float
                    Time shift in seconds.

            Example:
                >>> from ketos.audio_processing.annotation import AnnotationHandler
                >>> 
                >>> labels = 1
                >>> boxes = [10.0, 12.2]
                >>> handler = AnnotationHandler(labels, boxes)
                >>> handler._shift_annotations(2.0)
                >>> print(handler.boxes)
                [[12.0, 14.2, 0, inf]]
        """
        for b in self.boxes:
            b[0] += delay
            b[1] += delay
            for v in b: 
                v = round(v, self.precision)
        
    def _scale_annotations(self, scale=0):
        """ Scale the time axis by a constant factor.
            
            Args:
                scale: float
                    Scaling factor.

            Example:
                >>> from ketos.audio_processing.annotation import AnnotationHandler
                >>> 
                >>> labels = 1
                >>> boxes = [10.0, 12.2]
                >>> handler = AnnotationHandler(labels, boxes)
                >>> handler._scale_annotations(2.0)
                >>> print(handler.boxes)
                [[20.0, 24.4, 0, inf]]
        """
        for b in self.boxes:
            b[0] *= scale
            b[1] *= scale
            for v in b: 
                v = round(v, self.precision)



from setuptools import setup, find_packages

#
# create distribution and upload to pypi.org with:
#  
#   $ python setup.py sdist bdist_wheel
#   $ twine upload dist/*
#

setup(name='ketos',
      version='1.1.2',
      description="Python package for developing deep-learning-based models for the detection and classification of underwater sounds",
      # TODO: define a function readme() that reads the contents of a readme file
      # long_description=readme(),
      url='https://gitlab.meridian.cs.dal.ca/data_analytics_dal/packages/ketos',
      author='Oliver Kirsebom, Fabio Frazao',
      author_email='oliver.kirsebom@dal.ca, fsfrazao@dal.ca',
      license='GNU General Public License v3.0',
      packages=find_packages(),
      install_requires=[
          'numpy',
          'tables',
          'scipy',
          'pandas',
          'tensorflow',
          'scikit-learn',
          'scikit-image',
          'librosa',
          'datetime_glob',
          'matplotlib',
          'tqdm',
          'pint',
          'psutil',
          ],
      setup_requires=['pytest-runner', ],
      tests_require=['pytest', ],
      include_package_data=True,
      zip_safe=False)



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'audio' module within the ketos library
"""

import pytest
import ketos.audio_processing.audio as aud
import datetime
import numpy as np



def test_time_stamped_audio_signal_has_correct_begin_and_end_times(sine_audio):
    audio = sine_audio
    today = datetime.datetime.today()
    audio.time_stamp = today 
    seconds = sine_audio.duration()
    duration = datetime.timedelta(seconds=seconds)
    assert audio.begin() == today
    assert audio.end() == today + duration

def test_crop_audio_signal(sine_audio):
    audio = sine_audio
    seconds = len(audio.data) / audio.rate
    crop_begin = audio.begin() + datetime.timedelta(seconds=seconds/10.)
    crop_end = audio.end() - datetime.timedelta(seconds=seconds/10.)
    audio_cropped = audio
    audio_cropped.crop(begin=crop_begin, end=crop_end)
    seconds_cropped = len(audio_cropped.data) / audio_cropped.rate
    assert seconds_cropped/seconds == pytest.approx(8./10., rel=1./audio.rate)
    assert audio_cropped.begin() == crop_begin


def test_split_with_positive_sample_id(sine_audio):
    audio = sine_audio 
    t0 = audio.time_stamp
    first = audio.data[0]
    last = audio.data[-1]
    n = len(audio.data)
    m = int(0.1*n)
    dt = m / audio.rate
    s = audio.split(m)
    assert len(s.data) == m
    assert len(audio.data) == n-m
    assert first == s.data[0]
    assert last == audio.data[-1]
    assert audio.time_stamp == t0 + datetime.timedelta(microseconds=1e6*dt)
    assert s.time_stamp == t0

def test_split_with_sample_larger_than_length(sine_audio):
    audio = sine_audio
    n = len(audio.data)
    m = int(1.5*n)
    s = audio.split(m)
    assert len(s.data) == n
    assert audio.empty() == True

def test_split_with_negative_sample_id(sine_audio):
    audio = sine_audio
    t0 = audio.begin()
    t1 = audio.end()
    first = audio.data[0]
    last = audio.data[-1]
    n = len(audio.data)
    m = -int(0.2*n)
    dt = -m / audio.rate
    s = audio.split(m)
    assert len(s.data) == -m
    assert len(audio.data) == n+m
    assert first == audio.data[0]
    assert last == s.data[-1]
    assert s.time_stamp == t1 - datetime.timedelta(microseconds=1e6*dt)
    assert audio.time_stamp == t0

def test_append_audio_signal_to_itself(sine_audio):
    audio = sine_audio
    len_sum = 2 * len(audio.data)
    audio_copy = audio.copy()
    audio_copy.time_stamp = audio.end()
    audio.append(audio_copy)
    assert len(audio.data) == len_sum
    
def test_segment_audio_signal(sine_audio):
    lentot = sine_audio.duration()
    segs = sine_audio.segment(lentot/3)
    assert len(segs) == 3
    # recover original full length signal by concatenating segments
    v = segs[0].data
    for s in segs[1:]:
        v = np.concatenate((v, s.data))
    # check that the recovered audio signal matches the original
    assert np.all(v == sine_audio.data)
    # check behaviour when length is larger than 0.5 of total length
    segs = sine_audio.segment(0.8*lentot)
    assert len(segs) == 1
    assert segs[0].duration() == pytest.approx(0.8*lentot, abs=0.01)

def test_segment_audio_signal_w_annotations(sine_audio):
    lentot = sine_audio.duration()    
    # add some annotations
    t1 = 0.1 * lentot
    t2 = 0.2 * lentot
    sine_audio.annotate(labels=1, boxes=[t1, t2])
    t1 = 0.4 * lentot
    t2 = 0.6 * lentot
    sine_audio.annotate(labels=1, boxes=[t1, t2])
    t1 = 0.45 * lentot
    t2 = 0.95 * lentot
    sine_audio.annotate(labels=1, boxes=[t1, t2])
    # segment
    segs = sine_audio.segment(length=lentot/3, keep_time=True)
    assert len(segs) == 3
    # 1st segment should have one annotation
    s = segs[0]
    assert len(s.labels) == 1
    assert s.boxes[0][1] == 0.2 * lentot
    # 2nd segment should have two annotations
    s = segs[1]
    assert len(s.labels) == 2
    assert s.boxes[0][0] == 0.4 * lentot
    assert s.boxes[1][1] == 0.95 * lentot
    # 3rd segment should have one annotation
    s = segs[2]
    assert len(s.labels) == 1
    assert s.boxes[0][0] == 0.45 * lentot
    assert s.boxes[0][1] == 0.95 * lentot

def test_append_audio_signal_without_time_stamp_to_itself(sine_audio_without_time_stamp): 
    len_sum = 2 * len(sine_audio_without_time_stamp.data)
    sine_audio_without_time_stamp.append(sine_audio_without_time_stamp)
    assert len(sine_audio_without_time_stamp.data) == len_sum

def test_append_with_smoothing(sine_audio):
    audio = sine_audio
    t = audio.duration()
    at = audio.append(signal=audio, n_smooth=100)
    assert audio.duration() == pytest.approx(2.*t - 100/audio.rate, rel=1./audio.rate)
    assert at == audio.begin() + datetime.timedelta(microseconds=1e6*(t - 100/audio.rate))

def test_append_with_delay(sine_audio):
    audio = sine_audio
    t = audio.duration()
    delay = 3.0
    audio.append(signal=audio, delay=3)
    assert audio.duration() == 2.*t + 3.0

def test_append_with_max_length(sine_audio):
    audio = sine_audio
    audio2 = audio.copy()
    t = audio.duration()
    n = len(audio.data)
    nmax = int(1.5 * n)
    audio.append(signal=audio2, max_length=nmax)
    assert len(audio.data) == nmax
    assert len(audio2.data) == 2*n - nmax

def test_append_with_max_length_and_smooth(sine_audio):
    audio = sine_audio
    audio2 = audio.copy()
    t = audio.duration()
    n = len(audio.data)
    nmax = int(1.5 * n)
    n_smooth = 200
    audio.append(signal=audio2, n_smooth=n_smooth, max_length=nmax)
    assert len(audio.data) == nmax
    assert len(audio2.data) == 2*n - n_smooth - nmax

def test_append_with_max_length_and_delay(sine_audio):
    audio = sine_audio
    audio2 = audio.copy()
    t = audio.duration()
    n = len(audio.data)
    nmax = int(1.5 * n)
    delay = t
    audio.append(signal=audio2, delay=t, max_length=nmax)
    assert len(audio.data) == nmax
    assert len(audio2.data) == n
    assert np.ma.is_masked(audio.data[-1]) 

def test_append_delay_determined_from_time_stamps(sine_audio):
    audio = sine_audio
    audio2 = audio.copy()
    dt = 5.
    audio2.time_stamp = audio.end() + datetime.timedelta(microseconds=1e6*dt)
    t = audio.duration()
    n = len(audio.data)
    audio.append(signal=audio2)
    assert audio.duration() == 2*t + dt
    assert np.ma.is_masked(audio.data[n]) 
    
def test_add_identical_audio_signals(sine_audio):
    audio = sine_audio
    t = audio.duration()
    v = np.copy(audio.data)
    audio.add(signal=audio)
    assert audio.duration() == t
    assert np.all(audio.data == 2*v)
    
def test_add_identical_audio_signals_with_delay(sine_audio):
    audio = sine_audio
    t = audio.duration()
    v = np.copy(audio.data)
    delay = 1
    audio.add(signal=audio, delay=delay)
    assert audio.duration() == t
    i = int(audio.rate * delay)
    assert audio.data[5] == v[5]
    assert audio.data[i+5] == v[i+5] + v[5]    
    
def test_add_identical_audio_signals_with_scaling_factor(sine_audio):
    audio = sine_audio
    v = np.copy(audio.data)
    audio.add(signal=audio, scale=0.5)
    assert np.all(audio.data == 1.5*v)

def test_morlet_with_default_params():
    mor = aud.AudioSignal.morlet(rate=4000, frequency=20, width=1)
    assert len(mor.data) == int(6*1*4000) # check number of samples
    assert max(mor.data) == pytest.approx(1, abs=0.01) # check max signal is 1
    assert np.argmax(mor.data) == pytest.approx(0.5*len(mor.data), abs=1) # check peak is centered
    assert mor.data[0] == pytest.approx(0, abs=0.02) # check signal is approx zero at start

def test_gaussian_noise():
    noise = aud.AudioSignal.gaussian_noise(rate=2000, sigma=2, samples=40000)
    assert noise.std() == pytest.approx(2, rel=0.05) # check standard deviation
    assert noise.average() == pytest.approx(0, abs=6*2/np.sqrt(40000)) # check mean
    assert noise.duration() == 20 # check length

def test_clip(sine_audio):
    audio = sine_audio
    segs = audio.clip(boxes=[[0.1, 0.4],[0.3, 0.7]])
    assert len(segs) == 2
    assert segs[0].duration() == pytest.approx(0.3, abs=2./audio.rate)
    assert segs[1].duration() == pytest.approx(0.4, abs=2./audio.rate)
    assert audio.duration() == pytest.approx(3.0-0.6, abs=2./audio.rate)

@pytest.mark.test_resample
def test_resampled_signal_has_correct_rate(sine_wave_file):
    signal = aud.AudioSignal.from_wav(sine_wave_file)

    new_signal = signal.copy()
    new_signal.resample(new_rate=22000)
    assert new_signal.rate == 22000

    new_signal = signal.copy()
    new_signal.resample(new_rate=2000)
    assert new_signal.rate == 2000

@pytest.mark.test_resample
def test_resampled_signal_has_correct_length(sine_wave_file):
    signal = aud.AudioSignal.from_wav(sine_wave_file)

    duration = signal.duration()

    new_signal = signal.copy()
    new_signal.resample(new_rate=22000)
    assert len(new_signal.data) == duration * new_signal.rate 

    new_signal = signal.copy()
    new_signal.resample(new_rate=2000)
    assert len(new_signal.data) == duration * new_signal.rate 

@pytest.mark.test_resample
def test_resampling_preserves_signal_shape(const_wave_file):
    signal = aud.AudioSignal.from_wav(const_wave_file)
    new_signal = signal.copy()
    new_signal.resample(new_rate=22000)

    n = min(len(signal.data), len(new_signal.data))
    for i in range(n):
        assert signal.data[i] == new_signal.data[i]

@pytest.mark.test_resample
def test_resampling_preserves_signal_frequency(sine_wave_file):
    signal = aud.AudioSignal.from_wav(sine_wave_file)
    rate = signal.rate
    sig = signal.data
    y = abs(np.fft.rfft(sig))
    freq = np.argmax(y)
    freqHz = freq * rate / len(sig)
    signal = aud.AudioSignal(rate, sig)
    new_signal = signal.copy()
    new_signal.resample(new_rate=22000)
    new_y = abs(np.fft.rfft(new_signal.data))
    new_freq = np.argmax(new_y)
    new_freqHz = new_freq * new_signal.rate / len(new_signal.data)

    assert freqHz == new_freqHz

@pytest.mark.test_make_frames
def test_signal_is_padded(sine_wave):
    rate, sig = sine_wave
    duration = len(sig) / rate
    winlen = 2*duration
    winstep = 2*duration
    signal = aud.AudioSignal(rate, sig)
    frames = signal.make_frames(winlen=winlen, winstep=winstep, zero_padding=True)
    assert frames.shape[0] == 1
    assert frames.shape[1] == 2*len(sig)
    assert frames[0, len(sig)] == 0
    assert frames[0, 2*len(sig)-1] == 0

@pytest.mark.test_make_frames
def test_can_make_overlapping_frames(sine_wave):
    rate, sig = sine_wave
    duration = len(sig) / rate
    winlen = duration/2
    winstep = duration/4
    signal = aud.AudioSignal(rate, sig)
    frames = signal.make_frames(winlen=winlen, winstep=winstep)
    assert frames.shape[0] == 3
    assert frames.shape[1] == len(sig)/2

@pytest.mark.test_make_frames
def test_can_make_non_overlapping_frames(sine_wave):
    rate, sig = sine_wave
    duration = len(sig) / rate
    winlen = duration/4
    winstep = duration/2
    signal = aud.AudioSignal(rate, sig)
    frames = signal.make_frames(winlen, winstep)
    assert frames.shape[0] == 2
    assert frames.shape[1] == len(sig)/4

@pytest.mark.test_make_frames
def test_first_frame_matches_original_signal(sine_wave):
    rate, sig = sine_wave
    duration = len(sig) / rate
    winlen = duration/4
    winstep = duration/10
    signal = aud.AudioSignal(rate, sig)
    frames = signal.make_frames(winlen, winstep)
    assert frames.shape[0] == 8
    for i in range(int(winlen*rate)):
        assert sig[i] == pytest.approx(frames[0,i], rel=1E-6)

@pytest.mark.test_make_frames
def test_window_length_can_exceed_duration(sine_wave):
    rate, sig = sine_wave
    duration = len(sig) / rate
    winlen = 2 * duration
    winstep = duration
    signal = aud.AudioSignal(rate, sig)
    frames = signal.make_frames(winlen, winstep)
    assert frames.shape[0] == 1




"""This program cleans the temporary files created by tests. """

import shutil
import os

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_tmp = os.path.join(current_dir, 'assets/tmp')

if os.path.exists(path_to_tmp):
    shutil.rmtree(path_to_tmp)
os.makedirs(path_to_tmp)